#!/usr/bin/env python
"""A polars-to-polars transformation function for filtering subjects by sequence length."""
import logging
from collections.abc import Callable
from functools import partial

import hydra
import polars as pl
from omegaconf import DictConfig

logger = logging.getLogger(__name__)

from MEDS_transforms import PREPROCESS_CONFIG_YAML
from MEDS_transforms.mapreduce.mapper import map_over


def filter_subjects_by_num_measurements(df: pl.LazyFrame, min_measurements_per_subject: int) -> pl.LazyFrame:
    """Filters subjects by the number of dynamic (timestamp non-null) measurements they have.

    Args:
        df: The input DataFrame.
        min_measurements_per_subject: The minimum number of measurements a subject must have to be included.

    Returns:
        The filtered DataFrame.

    Examples:
        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 2, 2, 3, 3,    4],
        ...     "time":       [1, 2, 1, 1, 2, 1, None, None],
        ... })
        >>> filter_subjects_by_num_measurements(df, 1)
        shape: (7, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 1          ┆ 1    │
        │ 1          ┆ 2    │
        │ 1          ┆ 1    │
        │ 2          ┆ 1    │
        │ 2          ┆ 2    │
        │ 3          ┆ 1    │
        │ 3          ┆ null │
        └────────────┴──────┘
        >>> filter_subjects_by_num_measurements(df, 2)
        shape: (5, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 1          ┆ 1    │
        │ 1          ┆ 2    │
        │ 1          ┆ 1    │
        │ 2          ┆ 1    │
        │ 2          ┆ 2    │
        └────────────┴──────┘
        >>> filter_subjects_by_num_measurements(df, 3)
        shape: (3, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 1          ┆ 1    │
        │ 1          ┆ 2    │
        │ 1          ┆ 1    │
        └────────────┴──────┘
        >>> filter_subjects_by_num_measurements(df, 4)
        shape: (0, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        └────────────┴──────┘
        >>> filter_subjects_by_num_measurements(df, 2.2)
        Traceback (most recent call last):
            ...
        TypeError: min_measurements_per_subject must be an integer; got <class 'float'> 2.2
    """
    if not isinstance(min_measurements_per_subject, int):
        raise TypeError(
            f"min_measurements_per_subject must be an integer; got {type(min_measurements_per_subject)} "
            f"{min_measurements_per_subject}"
        )

    return df.filter(pl.col("time").count().over("subject_id") >= min_measurements_per_subject)


def filter_subjects_by_num_events(df: pl.LazyFrame, min_events_per_subject: int) -> pl.LazyFrame:
    """Filters subjects by the number of events (unique timepoints) they have.

    Args:
        df: The input DataFrame.
        min_events_per_subject: The minimum number of events a subject must have to be included.

    Returns:
        The filtered DataFrame.

    Examples:
        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4],
        ...     "time": [1, 1, 1, 1, 2, 1, 1, 2, 3, None, None, 1, 2, 3],
        ... })
        >>> with pl.Config(tbl_rows=15):
        ...     filter_subjects_by_num_events(df, 1)
        shape: (14, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 1          ┆ 1    │
        │ 1          ┆ 1    │
        │ 1          ┆ 1    │
        │ 2          ┆ 1    │
        │ 2          ┆ 2    │
        │ 2          ┆ 1    │
        │ 3          ┆ 1    │
        │ 3          ┆ 2    │
        │ 3          ┆ 3    │
        │ 4          ┆ null │
        │ 4          ┆ null │
        │ 4          ┆ 1    │
        │ 4          ┆ 2    │
        │ 4          ┆ 3    │
        └────────────┴──────┘
        >>> with pl.Config(tbl_rows=15):
        ...     filter_subjects_by_num_events(df, 2)
        shape: (11, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 2          ┆ 1    │
        │ 2          ┆ 2    │
        │ 2          ┆ 1    │
        │ 3          ┆ 1    │
        │ 3          ┆ 2    │
        │ 3          ┆ 3    │
        │ 4          ┆ null │
        │ 4          ┆ null │
        │ 4          ┆ 1    │
        │ 4          ┆ 2    │
        │ 4          ┆ 3    │
        └────────────┴──────┘
        >>> filter_subjects_by_num_events(df, 3)
        shape: (8, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 3          ┆ 1    │
        │ 3          ┆ 2    │
        │ 3          ┆ 3    │
        │ 4          ┆ null │
        │ 4          ┆ null │
        │ 4          ┆ 1    │
        │ 4          ┆ 2    │
        │ 4          ┆ 3    │
        └────────────┴──────┘
        >>> filter_subjects_by_num_events(df, 4)
        shape: (5, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 4          ┆ null │
        │ 4          ┆ null │
        │ 4          ┆ 1    │
        │ 4          ┆ 2    │
        │ 4          ┆ 3    │
        └────────────┴──────┘
        >>> filter_subjects_by_num_events(df, 5)
        shape: (0, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        └────────────┴──────┘
        >>> filter_subjects_by_num_events(df, 2.2)
        Traceback (most recent call last):
            ...
        TypeError: min_events_per_subject must be an integer; got <class 'float'> 2.2
    """
    if not isinstance(min_events_per_subject, int):
        raise TypeError(
            f"min_events_per_subject must be an integer; got {type(min_events_per_subject)} "
            f"{min_events_per_subject}"
        )

    return df.filter(pl.col("time").n_unique().over("subject_id") >= min_events_per_subject)


def filter_subjects_fntr(stage_cfg: DictConfig) -> Callable[[pl.LazyFrame], pl.LazyFrame]:
    """Returns a function that filters subjects by the number of measurements and events they have.

    Args:
        stage_cfg: The stage configuration. Arguments include: min_measurements_per_subject,
            min_events_per_subject, both of which should be integers or None which specify the minimum number
            of measurements and events a subject must have to be included, respectively.

    Returns:
        The function that filters subjects by the number of measurements and/or events they have.

    Examples:
        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 1, 1, 2, 2, 2, 3,    3,    4, 4, 4, 4,    5, 5, 5, 5],
        ...     "time":       [1, 1, 1, 1, 1, 1, 2, 3, None, None, 1, 2, 2, None, 1, 2, 3, 1],
        ... })
        >>> stage_cfg = DictConfig({"min_measurements_per_subject": 4, "min_events_per_subject": 2})
        >>> filter_subjects_fntr(stage_cfg)(df)
        shape: (4, 2)
        ┌────────────┬──────┐
        │ subject_id ┆ time │
        │ ---        ┆ ---  │
        │ i64        ┆ i64  │
        ╞════════════╪══════╡
        │ 5          ┆ 1    │
        │ 5          ┆ 2    │
        │ 5          ┆ 3    │
        │ 5          ┆ 1    │
        └────────────┴──────┘
    """
    compute_fns = []
    if stage_cfg.min_measurements_per_subject:
        logger.info(
            f"Filtering subjects with fewer than {stage_cfg.min_measurements_per_subject} measurements "
            "(observations of any kind)."
        )
        compute_fns.append(
            partial(
                filter_subjects_by_num_measurements,
                min_measurements_per_subject=stage_cfg.min_measurements_per_subject,
            )
        )
    if stage_cfg.min_events_per_subject:
        logger.info(
            f"Filtering subjects with fewer than {stage_cfg.min_events_per_subject} events "
            "(unique timepoints)."
        )
        compute_fns.append(
            partial(filter_subjects_by_num_events, min_events_per_subject=stage_cfg.min_events_per_subject)
        )

    def fn(data: pl.LazyFrame) -> pl.LazyFrame:
        for compute_fn in compute_fns:
            data = compute_fn(data)
        return data

    return fn


@hydra.main(
    version_base=None, config_path=str(PREPROCESS_CONFIG_YAML.parent), config_name=PREPROCESS_CONFIG_YAML.stem
)
def main(cfg: DictConfig):
    """TODO."""

    map_over(cfg, compute_fn=filter_subjects_fntr)


if __name__ == "__main__":  # pragma: no cover
    main()



#!/usr/bin/env python
import gzip
import logging
import warnings
from collections.abc import Callable
from enum import StrEnum
from pathlib import Path
from typing import TypeVar

import polars as pl

logger = logging.getLogger(__name__)


class SupportedFileFormats(StrEnum):
    """The supported file formats for dataframes we can read in, in priority order.

    The values of the enum are the allowed file suffix for the format.
    """

    PARQUET = ".parquet"
    CSV_GZ = ".csv.gz"
    CSV = ".csv"


def scan_csv_gz(fp: Path, **kwargs) -> pl.LazyFrame:
    with gzip.open(fp, mode="rb") as f:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            return pl.read_csv(f, **kwargs).lazy()


READERS = {
    SupportedFileFormats.PARQUET: pl.scan_parquet,
    SupportedFileFormats.CSV_GZ: scan_csv_gz,
    SupportedFileFormats.CSV: pl.scan_csv,
}


DF_T = TypeVar("DF_T")


def get_supported_fp(root_dir: Path, file_prefix: str | Path) -> tuple[Path, Callable[[Path], DF_T]]:
    """This function finds the best file path to read for a given root_dir and prefix.

    Args:
        root_dir: The root directory to search for files.
        file_prefix: The file prefix to search for.

    Raises:
        FileNotFoundError: If no files are found with the given prefix and an allowed suffix.

    Returns:
        The filepath with the matching prefix and the most preferred allowed suffix and an appropriate reader
        function for that file type.

    Examples:
        >>> from tempfile import TemporaryDirectory
        >>> df = pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}, schema={"a": pl.UInt8, "b": pl.Int64})
        >>> with TemporaryDirectory() as tmpdir:
        ...     tmpdir = Path(tmpdir)
        ...     fp = tmpdir / "test.csv"
        ...     df.write_csv(fp)
        ...     fp, reader = get_supported_fp(tmpdir, "test")
        ...     print(str(fp.relative_to(tmpdir)), reader(fp).collect())
        test.csv shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 1   ┆ 4   │
        │ 2   ┆ 5   │
        │ 3   ┆ 6   │
        └─────┴─────┘
        >>> with TemporaryDirectory() as tmpdir:
        ...     tmpdir = Path(tmpdir)
        ...     fp = tmpdir / "test.parquet"
        ...     csv_fp = tmpdir / "test.csv"
        ...     df.write_parquet(fp)
        ...     df.write_csv(csv_fp)
        ...     fp, reader = get_supported_fp(tmpdir, "test")
        ...     print(str(fp.relative_to(tmpdir)), reader(fp).collect())
        test.parquet shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ u8  ┆ i64 │
        ╞═════╪═════╡
        │ 1   ┆ 4   │
        │ 2   ┆ 5   │
        │ 3   ┆ 6   │
        └─────┴─────┘
        >>> import gzip
        >>> with TemporaryDirectory() as tmpdir:
        ...     tmpdir = Path(tmpdir)
        ...     fp = tmpdir / "test.csv.gz"
        ...     with gzip.open(fp, mode="wb") as f:
        ...         with warnings.catch_warnings():
        ...             warnings.simplefilter("ignore", category=UserWarning)
        ...             df.write_csv(f)
        ...     fp, reader = get_supported_fp(tmpdir, "test")
        ...     print(str(fp.relative_to(tmpdir)), reader(fp).collect())
        test.csv.gz shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 1   ┆ 4   │
        │ 2   ┆ 5   │
        │ 3   ┆ 6   │
        └─────┴─────┘
        >>> with TemporaryDirectory() as tmpdir:
        ...     tmpdir = Path(tmpdir)
        ...     fp = tmpdir / "test.json"
        ...     df.write_json(fp)
        ...     get_supported_fp(tmpdir, "test") # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        FileNotFoundError: No files found with prefix: test and allowed suffixes
            ['.parquet', '.csv.gz', '.csv']...
    """

    for suffix in list(SupportedFileFormats):
        fp = root_dir / f"{file_prefix}{suffix.value}"
        if fp.exists():
            logger.debug(f"Found file: {str(fp.resolve())}")
            return fp, READERS[suffix]
    raise FileNotFoundError(
        f"No files found with prefix: {file_prefix} and allowed suffixes "
        f"{[x.value for x in SupportedFileFormats]} in root dir {str(root_dir.resolve())}"
    )



"""Module for code that provides a structured DSL to specify columns of dataframes or operations on said.

This module contains two key concepts: column expressions and matchers.

Matchers are used to specify conditionals over dataframes. They are expressed simply as dictionaries mapping
column names to values. Exact equality is used to match the column values.

Column expressions currently support the following types:
  - COL (`'col'`): A column expression that extracts a specified column.
  - STR (`'str'`): A column expression that is a string, with interpolation allowed to other column names via
    python's f-string syntax.
  - LITERAL (`'literal'`): A column expression that is a literal value regardless of type. No interpolation is
    allowed here.

Column expressions can be expressed either dictionary or via a shorthand string. If a structured dictionary,
the dictionary has length 1 and the key is one of the column expression types and the value is the expression
target (e.g., the column to load for `COL`, the string to interpolate with `{...}` escaped interpolation
targets for `STR`, or the literal value for `LITERAL`). If a string, the string is interpreted as a `COL` if
it has no interpolation targets, and as a `STR` otherwise.

These types can be combined or filtered via two modes:
  - Coalescing: Multiple column expressions can be combined into a single expression where the first non-null
    is used by specifying them in an ordered list.
  - Conditional: A column expression can be conditionally applied to a dataframe based on a matcher, by
    specifying the column expression and matcher in a dictionary, in one of two possible forms:
    - A single key-value pair, where the key is a string realization of either a `COL` or `STR` type expressed
      as a string and the value is the matcher dictionary.
    - Two key-value pairs, where the first key is `"output"` and the value is the column expression and the
      second key is `"matcher"` and the value is the matcher dictionary.
"""

from __future__ import annotations

import re
from enum import StrEnum
from typing import Annotated, Any

import polars as pl
from omegaconf import DictConfig, ListConfig, OmegaConf


def is_matcher(matcher_cfg: dict[str, Any]) -> tuple[bool, str | None]:
    """Checks if a dictionary is a valid matcher configuration.

    Args:
        matcher_cfg: A dictionary of key-value pairs to match against.

    Returns:
        bool: True if the input is a valid matcher configuration, False otherwise.

    Examples:
        >>> is_matcher({"foo": "bar"})
        (True, None)
        >>> is_matcher(DictConfig({"foo": "bar"}))
        (True, None)
        >>> is_matcher(["foo", "bar"])
        (False, "Matcher configuration must be a dictionary. Got ['foo', 'bar']")
        >>> is_matcher({"foo": "bar", 32: "baz"})
        (False, "Matcher configuration must be a dictionary with string keys. Got {'foo': 'bar', 32: 'baz'}")
        >>> is_matcher({"foo": {"regex": "bar"}})
        (True, None)
        >>> is_matcher({"foo": {"regex": "bar", "group_index": 0}}) # doctest: +NORMALIZE_WHITESPACE
        (False, "If the matcher spec is a dictionary, it must have only a 'regex' key.
                 Got {'regex': 'bar', 'group_index': 0} for foo")
        >>> is_matcher({})
        (True, None)
    """
    if not isinstance(matcher_cfg, (dict, DictConfig)):
        return False, f"Matcher configuration must be a dictionary. Got {matcher_cfg}"
    if not all(isinstance(k, str) for k in matcher_cfg.keys()):
        return False, f"Matcher configuration must be a dictionary with string keys. Got {matcher_cfg}"
    for k, cfg in matcher_cfg.items():
        if isinstance(cfg, (dict, DictConfig)) and set(cfg.keys()) != {
            "regex",
        }:
            return False, (
                f"If the matcher spec is a dictionary, it must have only a 'regex' key. Got {cfg} for {k}"
            )

    return True, None


def matcher_to_expr(matcher_cfg: DictConfig | dict) -> tuple[pl.Expr, set[str]]:
    """Returns an expression and the necessary columns to match a collection of key-value pairs.

    Currently, this only supports checking for equality between column names and values.
    TODO: Expand (as needed only) to support other types of matchers.

    Args:
        matcher_cfg: A dictionary of key-value pairs to match against.

    Raises:
        ValueError: If the matcher configuration is not a dictionary.

    Returns:
        pl.Expr: A Polars expression that matches the key-value pairs in the input dictionary.
        set[str]: The set of input columns needed to form the returned expression.

    Examples:
        >>> expr, cols = matcher_to_expr({"foo": "bar", "buzz": "baz"})
        >>> print(expr)
        [(col("foo")) == (String(bar))].all_horizontal([[(col("buzz")) == (String(baz))]])
        >>> sorted(cols)
        ['buzz', 'foo']
        >>> expr, cols = matcher_to_expr(DictConfig({"foo": "bar", "buzz": "baz"}))
        >>> print(expr)
        [(col("foo")) == (String(bar))].all_horizontal([[(col("buzz")) == (String(baz))]])
        >>> sorted(cols)
        ['buzz', 'foo']
        >>> expr, cols = matcher_to_expr(DictConfig({"foo": "bar", "buzz": {"regex": "baz"}}))
        >>> print(expr)
        [(col("foo")) == (String(bar))].all_horizontal([col("buzz").str.contains([String(baz)])])
        >>> sorted(cols)
        ['buzz', 'foo']
        >>> matcher_to_expr(["foo", "bar"])
        Traceback (most recent call last):
            ...
        ValueError: Matcher configuration must be a dictionary. Got ['foo', 'bar']
    """
    matcher_valid, matcher_errs = is_matcher(matcher_cfg)
    if not matcher_valid:
        raise ValueError(matcher_errs)

    all_exprs = []
    for k, v in matcher_cfg.items():
        if isinstance(v, (dict, DictConfig)):
            expr = pl.col(k).str.contains(v["regex"])
        else:
            expr = pl.col(k) == v
        all_exprs.append(expr)

    return pl.all_horizontal(all_exprs), set(matcher_cfg.keys())


STR_INTERPOLATION_REGEX = r"\{([^}]+)\}"


class ColExprType(StrEnum):
    """Enumeration of the different types of column expressions that can be parsed.

    Members:
        COL: A column expression that extracts a specified column.
        STR: A column expression that is a string, with interpolation allowed to other column names
            via python's f-string syntax.
        LITERAL: A column expression that is a literal value regardless of type. No interpolation is allowed
            here.
        EXTRACT: A column expression that extracts a substring from a column using a regex pattern.
    """

    COL = "col"
    STR = "str"
    LITERAL = "literal"
    EXTRACT = "extract"

    @classmethod
    def _is_valid_extract_cfg(cls, cfg: dict[Any, Any]) -> tuple[bool, str | None]:
        """Checks if a dictionary is a valid extract configuration.

        Args:
            cfg: A dictionary of extract configuration. This is the "inner part" of an overall column
                expression."

        Returns:
            bool: True if the input is a valid extract configuration, False otherwise.
            str | None: The reason the input is invalid, if it is invalid.

        Examples:
            >>> ColExprType._is_valid_extract_cfg({"from": "foo", "regex": "bar"})
            (True, None)
            >>> ColExprType._is_valid_extract_cfg({"from": "foo", "regex": "bar", "group_index": 0})
            (True, None)
            >>> ColExprType._is_valid_extract_cfg(32)
            (False, 'Extract expressions must be a dictionary. Got 32')
            >>> ColExprType._is_valid_extract_cfg({"from": "foo"})
            (False, "Extract expressions must have a 'from' and 'regex' key. Got ['from']")
            >>> ColExprType._is_valid_extract_cfg({"from": ["buzz"], "regex": "bar"})
            (False, "Extract expressions must have a <class 'str'> value for 'from'. Got ['buzz']")
            >>> ColExprType._is_valid_extract_cfg({"from": "foo", "regex": 32})
            (False, "Extract expressions must have a <class 'str'> value for 'regex'. Got 32")
            >>> ColExprType._is_valid_extract_cfg({"from": "foo", "regex": "bar", "group_index": "baz"})
            (False, "Extract expressions must have a non-negative integer value for 'group_index'. Got baz")
            >>> ColExprType._is_valid_extract_cfg({"from": "foo", "regex": "bar", "group_index": -1})
            (False, "Extract expressions must have a non-negative integer value for 'group_index'. Got -1")
            >>> ColExprType._is_valid_extract_cfg({"from": "foo", "regex": "bar", "ex": "baz"})
            (False, "Extract expressions must have only 'from', 'regex', and 'group_index' keys. Got ['ex']")
        """

        MANDATORY_KEYS = {"from": str, "regex": str}
        ALLOWED_KEYS = {**MANDATORY_KEYS, "group_index": Annotated[int, "non-negative integer"]}

        if not isinstance(cfg, (dict, DictConfig)):
            return False, f"Extract expressions must be a dictionary. Got {cfg}"
        if not set(MANDATORY_KEYS).issubset(set(cfg.keys())):
            return False, f"Extract expressions must have a 'from' and 'regex' key. Got {list(cfg.keys())}"
        if not set(ALLOWED_KEYS).issuperset(set(cfg.keys())):
            return False, (
                "Extract expressions must have only 'from', 'regex', and 'group_index' keys. "
                f"Got {sorted(list(set(cfg.keys()) - set(ALLOWED_KEYS)))}"
            )

        for key, allowed_T in ALLOWED_KEYS.items():
            if key not in cfg:
                continue
            val = cfg[key]

            if key == "group_index":
                if not isinstance(val, int) or val < 0:
                    return False, (
                        f"Extract expressions must have a non-negative integer value for '{key}'. Got {val}"
                    )
            else:
                if not isinstance(val, allowed_T):
                    return False, f"Extract expressions must have a {allowed_T} value for '{key}'. Got {val}"

        return True, None

    @classmethod
    def is_valid(cls, expr_dict: dict[ColExprType, Any]) -> tuple[bool, str | None]:
        """Checks if a dictionary of expression key to value is a valid column expression.

        Args:
            expr_dict: A dictionary of column expression type to value.

        Returns:
            bool: True if the input is a valid column expression, False otherwise.
            str | None: The reason the input is invalid, if it is invalid.

        Examples:
            >>> ColExprType.is_valid({"col": "foo"})
            (True, None)
            >>> ColExprType.is_valid({"col": 32})
            (False, 'Column expressions must have a string value. Got 32')
            >>> ColExprType.is_valid({ColExprType.STR: "bar//{foo}"})
            (True, None)
            >>> ColExprType.is_valid({ColExprType.STR: ["bar//{foo}"]})
            (False, "String interpolation expressions must have a string value. Got ['bar//{foo}']")
            >>> ColExprType.is_valid({"literal": ["baz", 32]})
            (True, None)
            >>> ColExprType.is_valid({"col": "foo", "str": "bar"}) # doctest: +NORMALIZE_WHITESPACE
            (False, "Column expressions can only contain a single key-value pair.
                    Got {'col': 'foo', 'str': 'bar'}")
            >>> ColExprType.is_valid({"foo": "bar"})
            (False, "Column expressions must have a key in ColExprType: ... Got foo")
            >>> ColExprType.is_valid([("col", "foo")])
            (False, "Column expressions must be a dictionary. Got [('col', 'foo')]")
        """

        if not isinstance(expr_dict, (dict, DictConfig)):
            return False, f"Column expressions must be a dictionary. Got {expr_dict}"
        if len(expr_dict) != 1:
            return False, f"Column expressions can only contain a single key-value pair. Got {expr_dict}"

        expr_type, expr_val = next(iter(expr_dict.items()))
        match expr_type:
            case cls.COL if isinstance(expr_val, str):
                return True, None
            case cls.COL:
                return False, f"Column expressions must have a string value. Got {expr_val}"
            case cls.STR if isinstance(expr_val, str):
                return True, None
            case cls.STR:
                return False, f"String interpolation expressions must have a string value. Got {expr_val}"
            case cls.LITERAL:
                return True, None
            case cls.EXTRACT:
                return cls._is_valid_extract_cfg(expr_val)
            case _:
                return (
                    False,
                    f"Column expressions must have a key in ColExprType: {[x.value for x in cls]}. Got "
                    f"{expr_type}",
                )

    @classmethod
    def to_pl_expr(cls, expr_type: ColExprType, expr_val: Any) -> tuple[pl.Expr, set[str]]:
        """Converts a column expression type and value to a Polars expression.

        Args:
            expr_type: The type of column expression.
            expr_val: The value of the column expression.

        Returns:
            pl.Expr: A Polars expression that extracts the column from the metadata DataFrame.
            set[str]: The set of input columns needed to form the returned expression.

        Raises:
            ValueError: If the column expression type is invalid.

        Examples:
            >>> print(*ColExprType.to_pl_expr(ColExprType.COL, "foo"))
            col("foo") {'foo'}
            >>> expr, cols = ColExprType.to_pl_expr(ColExprType.STR, "bar//{foo}//{baz}")
            >>> print(expr)
            String(bar//).str.concat_horizontal([col("foo"), String(//), col("baz")])
            >>> sorted(cols)
            ['baz', 'foo']
            >>> expr, cols = ColExprType.to_pl_expr(ColExprType.LITERAL, ListConfig(["foo", "bar"]))
            >>> print(expr)
            ["foo", "bar"]
            >>> cols
            set()
            >>> expr, cols = ColExprType.to_pl_expr(ColExprType.EXTRACT, {"from": "foo", "regex": "bar"})
            >>> print(expr)
            col("foo").str.extract([String(bar)])
            >>> sorted(cols)
            ['foo']
            >>> ColExprType.to_pl_expr(ColExprType.COL, 32)
            Traceback (most recent call last):
                ...
            ValueError: ...
        """
        is_valid, err_msg = cls.is_valid({expr_type: expr_val})
        if not is_valid:
            raise ValueError(err_msg)

        match expr_type:
            case cls.COL:
                return pl.col(expr_val), {expr_val}
            case cls.STR:
                cols = list(re.findall(STR_INTERPOLATION_REGEX, expr_val))
                expr_val = re.sub(STR_INTERPOLATION_REGEX, "{}", expr_val)
                return pl.format(expr_val, *cols), set(cols)
            case cls.LITERAL:
                if isinstance(expr_val, ListConfig):
                    expr_val = OmegaConf.to_object(expr_val)
                return pl.lit(expr_val), set()
            case cls.EXTRACT:
                expr = pl.col(expr_val["from"]).str.extract(
                    expr_val["regex"], group_index=expr_val.get("group_index", 1)
                )
                cols = {expr_val["from"]}
                return expr, cols


def parse_col_expr(cfg: str | list | dict[str, str] | ListConfig | DictConfig) -> dict:
    """Parses a column expression configuration object into a dictionary expressing the desired expression.

    Args:
        col_expr: A configuration object that specifies how to extract a column from the metadata. See the
            module docstring for formatting details.

    Returns:
        A dictionary specifying, in a structured form, the desired column expression.

    Examples:
        >>> parse_col_expr("foo")
        {'col': 'foo'}
        >>> parse_col_expr("bar//{foo}")
        {'str': 'bar//{foo}'}
        >>> parse_col_expr({'col': 'bar//{foo}'})
        {'col': 'bar//{foo}'}
        >>> parse_col_expr({"literal": ["foo", "bar"]})
        {'literal': ['foo', 'bar']}
        >>> parse_col_expr({"output": "foo", "matcher": {"bar": "baz"}})
        {'output': {'col': 'foo'}, 'matcher': {'bar': 'baz'}}
        >>> parse_col_expr({"output": "foo", "matcher": {32: "baz"}}) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: A pre-specified output/matcher configuration must have a valid matcher dictionary. Got
                    cfg['matcher']={32: 'baz'} which has errors: ...
        >>> parse_col_expr({"foo": {"bar": "baz"}})
        {'output': {'col': 'foo'}, 'matcher': {'bar': 'baz'}}
        >>> parse_col_expr({"foo": {32: "baz"}}) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: A simple-form conditional expression is expressed with a single key-value pair dict,
                    where the key is not a column expression type and the value is a valid matcher dict. This
                    config has a single key-value pair with key foo but an invalid matcher: {32: 'baz'} which
                    has errors: ...
        >>> parse_col_expr(["bar//{foo}", {"str": "bar//UNK"}])
        [{'str': 'bar//{foo}'}, {'str': 'bar//UNK'}]
        >>> parse_col_expr({"foo": "bar", "buzz": "baz", "fuzz": "fizz"}) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Dictionary column expression must either be explicit output/matcher configs, with two
                    keys, 'output' and 'matcher' with a valid matcher dictionary, or a simple column
                    expression with a single key-value pair where the key is a column expression type, or a
                    simple-form conditional expression with a single key-value pair where the key is the
                    conditional value and the value is a valid matcher dict. Got a dictionary with 3 elements:
                    {'foo': 'bar', 'buzz': 'baz', 'fuzz': 'fizz'}
        >>> parse_col_expr(('foo', 'bar')) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: A simple column expression must be a string, list, or dictionary.
                    Got <class 'tuple'>: ('foo', 'bar')
        >>> parse_col_expr({"col": "foo", "str": "bar"}) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Dictionary column expression must either be explicit output/matcher configs, with two
                    keys, 'output' and 'matcher' with a valid matcher dictionary, or a simple column
                    expression with a single key-value pair where the key is a column expression type, or a
                    simple-form conditional expression with a single key-value pair where the key is the
                    conditional value and the value is a valid matcher dict. Got a dictionary with 2 elements:
                    {'col': 'foo', 'str': 'bar'}
        >>> parse_col_expr(["foo", 32]) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: If a list (which coalesces columns), all elements must be strings or dictionaries.
                    Got: ['foo', 32]
    """
    match cfg:
        case str() if re.search(STR_INTERPOLATION_REGEX, cfg):
            return {"str": cfg}
        case str():
            return {"col": cfg}
        case list() | ListConfig() if all(isinstance(x, (str, dict, DictConfig)) for x in cfg):
            return [parse_col_expr(x) for x in cfg]
        case list() | ListConfig():
            raise ValueError(
                "If a list (which coalesces columns), all elements must be strings or dictionaries. "
                f"Got: {cfg}"
            )
        case dict() | DictConfig() if set(cfg.keys()) == {"output", "matcher"}:
            matcher_valid, matcher_errs = is_matcher(cfg["matcher"])
            if not matcher_valid:
                raise ValueError(
                    "A pre-specified output/matcher configuration must have a valid matcher dictionary. "
                    f"Got cfg['matcher']={cfg['matcher']} which has errors: {matcher_errs}"
                )
            return {"output": parse_col_expr(cfg["output"]), "matcher": cfg["matcher"]}
        case dict() | DictConfig() if len(cfg) == 1 and ColExprType.is_valid(cfg)[0]:
            return cfg
        case dict() | DictConfig() if len(cfg) == 1:
            out_cfg, matcher_cfg = next(iter(cfg.items()))
            matcher_valid, matcher_errs = is_matcher(matcher_cfg)
            if matcher_valid:
                return {"output": parse_col_expr(out_cfg), "matcher": matcher_cfg}
            else:
                raise ValueError(
                    "A simple-form conditional expression is expressed with a single key-value pair dict, "
                    "where the key is not a column expression type and the value is a valid matcher dict. "
                    f"This config has a single key-value pair with key {out_cfg} but an invalid matcher: "
                    f"{matcher_cfg} which has errors: {matcher_errs}"
                )
        case dict() | DictConfig():
            raise ValueError(
                "Dictionary column expression must either be explicit output/matcher configs, with two keys, "
                "'output' and 'matcher' with a valid matcher dictionary, or a simple column expression with "
                "a single key-value pair where the key is a column expression type, or a simple-form "
                "conditional expression with a single key-value pair where the key is the conditional value "
                f"and the value is a valid matcher dict. Got a dictionary with {len(cfg)} elements: {cfg}"
            )
        case _:
            raise ValueError(
                f"A simple column expression must be a string, list, or dictionary. Got {type(cfg)}: {cfg}"
            )


def structured_expr_to_pl(cfg: dict | list[dict] | ListConfig | DictConfig) -> tuple[pl.Expr, set[str]]:
    """Converts a structured column expression configuration object to a Polars expression.

    Args:
        structured_expr: A structured column expression configuration object. See the module docstring for DSL
            details.

    Returns:
        pl.Expr: A Polars expression that extracts the column from the metadata DataFrame.
        set[str]: The set of input columns needed to form the returned expression.

    Raises:
        ValueError: If the configuration object is invalid.

    Examples:
        >>> expr, cols = structured_expr_to_pl([{"col": "foo"}, {"str": "bar//{baz}"}, {"literal": "fizz"}])
        >>> print(expr)
        col("foo").coalesce([String(bar//).str.concat_horizontal([col("baz")]), String(fizz)])
        >>> sorted(cols)
        ['baz', 'foo']
        >>> expr, cols = structured_expr_to_pl({"output": {"literal": "foo"}, "matcher": {"bar": "baz"}})
        >>> print(expr)
        .when([(col("bar")) == (String(baz))].all_horizontal()).then(String(foo)).otherwise(null)
        >>> sorted(cols)
        ['bar']
        >>> expr, cols = structured_expr_to_pl({"extract": {"from": "foo", "regex": r"b/(.*)"}})
        >>> print(expr)
        col("foo").str.extract([String(b/(.*))])
        >>> sorted(cols)
        ['foo']
        >>> expr, cols = structured_expr_to_pl({"extract": {"from": "foo", "regex": "bar", "group_index": 0}})
        >>> print(expr)
        col("foo").str.extract([String(bar)])
        >>> sorted(cols)
        ['foo']
        >>> structured_expr_to_pl(["foo", 32])
        Traceback (most recent call last):
            ...
        ValueError: Error processing list config on field 1 for ['foo', 32]
        >>> structured_expr_to_pl({"output": 32, "matcher": {"bar": "baz"}}) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Error processing output/matcher config output expression for
                    {'output': 32, 'matcher': {'bar': 'baz'}}
        >>> structured_expr_to_pl({"output": "foo", "matcher": {32: "baz"}}) # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: A pre-specified output/matcher configuration must have a valid matcher dictionary. Got
                    cfg['matcher']={32: 'baz'} which has errors: ...
        >>> structured_expr_to_pl({"col": 32})
        Traceback (most recent call last):
            ...
        ValueError: Column expressions must have a string value. Got 32
        >>> structured_expr_to_pl("foo")
        Traceback (most recent call last):
            ...
        ValueError: A structured column expression must be a list or dictionary. Got <class 'str'>: foo
    """

    match cfg:
        case list() | ListConfig() as cfg_fields:
            component_exprs = []
            needed_cols = set()
            for i, field in enumerate(cfg_fields):
                try:
                    expr, cols = cfg_to_expr(field)
                except ValueError as e:
                    raise ValueError(f"Error processing list config on field {i} for {cfg}") from e
                component_exprs.append(expr)
                needed_cols.update(cols)
            return pl.coalesce(*component_exprs), needed_cols
        case dict() | DictConfig() if set(cfg.keys()) == {"output", "matcher"}:
            matcher_valid, matcher_errs = is_matcher(cfg["matcher"])
            if not matcher_valid:
                raise ValueError(
                    "A pre-specified output/matcher configuration must have a valid matcher dictionary. "
                    f"Got cfg['matcher']={cfg['matcher']} which has errors: {matcher_errs}"
                )

            matcher_expr, matcher_cols = matcher_to_expr(cfg["matcher"])
            try:
                out_expr, out_cols = cfg_to_expr(cfg["output"])
            except ValueError as e:
                raise ValueError(f"Error processing output/matcher config output expression for {cfg}") from e
            return pl.when(matcher_expr).then(out_expr), out_cols | matcher_cols

        case dict() | DictConfig() if ColExprType.is_valid(cfg)[0]:
            expr_type, expr_val = next(iter(cfg.items()))
            return ColExprType.to_pl_expr(expr_type, expr_val)
        case dict() | DictConfig():
            _, err_msg = ColExprType.is_valid(cfg)
            raise ValueError(err_msg)
        case _:
            raise ValueError(
                f"A structured column expression must be a list or dictionary. Got {type(cfg)}: {cfg}"
            )


def cfg_to_expr(cfg: str | ListConfig | DictConfig) -> tuple[pl.Expr, set[str]]:
    """Converts a metadata column configuration object to a Polars expression.

    Args:
        cfg: A configuration object that specifies how to extract a column from the metadata. See the module
            docstring for formatting details.

    Returns:
        pl.Expr: A Polars expression that extracts the column from the metadata DataFrame.
        set[str]: The set of input columns needed to form the returned expression.

    Examples:
        >>> data = pl.DataFrame({
        ...     "foo": ["a", "b", "c"],
        ...     "bar": ["d", "e", "f"],
        ...     "baz": [1,   2,   3]
        ... })
        >>> expr, cols = cfg_to_expr("foo")
        >>> data.select(expr.alias("out"))["out"].to_list()
        ['a', 'b', 'c']
        >>> sorted(cols)
        ['foo']
        >>> expr, cols = cfg_to_expr("bar//{foo}//{baz}")
        >>> data.select(expr.alias("out"))["out"].to_list()
        ['bar//a//1', 'bar//b//2', 'bar//c//3']
        >>> sorted(cols)
        ['baz', 'foo']
        >>> expr, cols = cfg_to_expr({"literal": 34.2})
        >>> data.select(expr.alias("out"))["out"].to_list()
        [34.2]
        >>> cols
        set()
        >>> expr, cols = cfg_to_expr({"{baz}//{bar}": {"foo": "a"}})
        >>> data.select(expr.alias("out"))["out"].to_list()
        ['1//d', None, None]
        >>> sorted(cols)
        ['bar', 'baz', 'foo']
        >>> expr, cols = cfg_to_expr({"extract": {"regex": "([ac]).*", "from": "foo"}})
        >>> data.select(expr.alias("out"))["out"].to_list()
        ['a', None, 'c']
        >>> sorted(cols)
        ['foo']
        >>> cfg = [
        ...    {"matcher": {"baz": 2}, "output": {"str": "bar//{baz}"}},
        ...    {"literal": "34.2"},
        ... ]
        >>> expr, cols = cfg_to_expr(cfg)
        >>> data.select(expr.alias("out"))["out"].to_list()
        ['34.2', 'bar//2', '34.2']
        >>> sorted(cols)
        ['baz']
    """
    structured_expr = parse_col_expr(cfg)
    return structured_expr_to_pl(structured_expr)



"""Tests the aggregate_code_metadata.py script.

Set the bash env variable `DO_USE_LOCAL_SCRIPTS=1` to use the local py files, rather than the installed
scripts.
"""

import polars as pl

from tests.MEDS_Transforms import AGGREGATE_CODE_METADATA_SCRIPT
from tests.MEDS_Transforms.transform_tester_base import (
    MEDS_CODE_METADATA_SCHEMA,
    MEDS_SHARDS,
    single_stage_transform_tester,
)

WANT_OUTPUT_CODE_METADATA_FILE = """
code,code/n_occurrences,code/n_subjects,values/n_occurrences,values/n_subjects,values/sum,values/sum_sqd,values/n_ints,values/min,values/max,description,parent_codes
,44,4,28,4,3198.8389005974336,382968.28937288234,6,86.0,175.271118,,
ADMISSION//CARDIAC,2,2,0,0,0,0,0,,,,
ADMISSION//ORTHOPEDIC,1,1,0,0,0,0,0,,,,
ADMISSION//PULMONARY,1,1,0,0,0,0,0,,,,
DISCHARGE,4,4,0,0,0,0,0,,,,
DOB,4,4,0,0,0,0,0,,,,
EYE_COLOR//BLUE,1,1,0,0,0,0,0,,,"Blue Eyes. Less common than brown.",
EYE_COLOR//BROWN,1,1,0,0,0,0,0,,,"Brown Eyes. The most common eye color.",
EYE_COLOR//HAZEL,2,2,0,0,0,0,0,,,"Hazel eyes. These are uncommon",
HEIGHT,4,4,4,4,656.8389005974336,108056.12937288235,0,156.485596,175.271118,,
HR,12,4,12,4,1360.5000000000002,158538.77,2,86.0,170.199997,"Heart Rate",LOINC/8867-4
TEMP,12,4,12,4,1181.4999999999998,116373.38999999998,4,95.5,100.400002,"Body Temperature",LOINC/8310-5
"""

WANT_OUTPUT_CODE_METADATA_FILE = pl.DataFrame(
    {
        "code": [
            None,
            "ADMISSION//CARDIAC",
            "ADMISSION//ORTHOPEDIC",
            "ADMISSION//PULMONARY",
            "DISCHARGE",
            "DOB",
            "EYE_COLOR//BLUE",
            "EYE_COLOR//BROWN",
            "EYE_COLOR//HAZEL",
            "HEIGHT",
            "HR",
            "TEMP",
        ],
        "code/n_occurrences": [44, 2, 1, 1, 4, 4, 1, 1, 2, 4, 12, 12],
        "code/n_subjects": [4, 2, 1, 1, 4, 4, 1, 1, 2, 4, 4, 4],
        "values/n_occurrences": [28, 0, 0, 0, 0, 0, 0, 0, 0, 4, 12, 12],
        "values/n_subjects": [4, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4],
        "values/sum": [
            3198.8389005974336,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            656.8389005974336,
            1360.5000000000002,
            1181.4999999999998,
        ],
        "values/sum_sqd": [
            382968.28937288234,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            108056.12937288235,
            158538.77,
            116373.38999999998,
        ],
        "values/n_ints": [6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 4],
        "values/min": [86.0, None, None, None, None, None, None, None, None, 156.485596, 86.0, 95.5],
        "values/max": [
            175.271118,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            175.271118,
            170.199997,
            100.400002,
        ],
        "values/quantiles": [
            {"values/quantile/0.25": 99.9, "values/quantile/0.5": 105.1, "values/quantile/0.75": 113.4},
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            {
                "values/quantile/0.25": 160.395311,
                "values/quantile/0.5": 164.686884,
                "values/quantile/0.75": 164.686884,
            },
            {"values/quantile/0.25": 107.5, "values/quantile/0.5": 112.5, "values/quantile/0.75": 113.4},
            {"values/quantile/0.25": 96.2, "values/quantile/0.5": 99.9, "values/quantile/0.75": 100.0},
        ],
        "description": [
            None,
            None,
            None,
            None,
            None,
            None,
            "Blue Eyes. Less common than brown.",
            "Brown Eyes. The most common eye color.",
            "Hazel eyes. These are uncommon",
            None,
            "Heart Rate",
            "Body Temperature",
        ],
        "parent_codes": [
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            ["LOINC/8867-4"],
            ["LOINC/8310-5"],
        ],
    },
    schema={
        **{k: v for k, v in MEDS_CODE_METADATA_SCHEMA.items() if k != "code/vocab_index"},
        "parent_codes": pl.List(pl.String),
        "values/quantiles": pl.Struct(
            {
                "values/quantile/0.25": pl.Float32,
                "values/quantile/0.5": pl.Float32,
                "values/quantile/0.75": pl.Float32,
            }
        ),
    },
)

MEDS_CODE_METADATA_FILE = """
code,description,parent_codes
EYE_COLOR//BLUE,"Blue Eyes. Less common than brown.",
EYE_COLOR//BROWN,"Brown Eyes. The most common eye color.",
EYE_COLOR//HAZEL,"Hazel eyes. These are uncommon",
HR,"Heart Rate",LOINC/8867-4
TEMP,"Body Temperature",LOINC/8310-5
"""

AGGREGATIONS = [
    "code/n_occurrences",
    "code/n_subjects",
    "values/n_occurrences",
    "values/n_subjects",
    "values/sum",
    "values/sum_sqd",
    "values/n_ints",
    "values/min",
    "values/max",
    {"name": "values/quantiles", "quantiles": [0.25, 0.5, 0.75]},
]


def test_aggregate_code_metadata():
    single_stage_transform_tester(
        transform_script=AGGREGATE_CODE_METADATA_SCRIPT,
        stage_name="aggregate_code_metadata",
        transform_stage_kwargs={"aggregations": AGGREGATIONS, "do_summarize_over_all_codes": True},
        want_metadata=WANT_OUTPUT_CODE_METADATA_FILE,
        input_code_metadata=MEDS_CODE_METADATA_FILE,
        do_use_config_yaml=True,
        assert_no_other_outputs=False,
        df_check_kwargs={"check_column_order": False},
    )

    # Test with shards re-mapped so it has to use the splits file.
    remapped_shards = {str(i): v for i, v in enumerate(MEDS_SHARDS.values())}
    single_stage_transform_tester(
        transform_script=AGGREGATE_CODE_METADATA_SCRIPT,
        stage_name="aggregate_code_metadata",
        transform_stage_kwargs={"aggregations": AGGREGATIONS, "do_summarize_over_all_codes": True},
        want_metadata=WANT_OUTPUT_CODE_METADATA_FILE,
        input_code_metadata=MEDS_CODE_METADATA_FILE,
        do_use_config_yaml=True,
        assert_no_other_outputs=False,
        df_check_kwargs={"check_column_order": False},
        input_shards=remapped_shards,
    )

    single_stage_transform_tester(
        transform_script=AGGREGATE_CODE_METADATA_SCRIPT,
        stage_name="aggregate_code_metadata",
        transform_stage_kwargs={"aggregations": AGGREGATIONS, "do_summarize_over_all_codes": True},
        want_metadata=WANT_OUTPUT_CODE_METADATA_FILE,
        input_code_metadata=MEDS_CODE_METADATA_FILE,
        do_use_config_yaml=True,
        input_shards=remapped_shards,
        splits_fp=None,
        should_error=True,
    )



"""Basic utilities for parallelizable map operations on sharded MEDS datasets with caching and locking."""

import inspect
import logging
from collections.abc import Callable, Generator
from datetime import datetime
from enum import Enum, StrEnum, auto
from functools import partial, wraps
from pathlib import Path
from typing import Any, NotRequired, TypedDict, TypeVar

import hydra
import polars as pl
from meds import subject_id_field, subject_splits_filepath
from omegaconf import DictConfig, ListConfig

from ..parser import is_matcher, matcher_to_expr
from ..utils import stage_init, write_lazyframe
from .utils import rwlock_wrap, shard_iterator

logger = logging.getLogger(__name__)

DF_T = TypeVar("DF_T")

COMPUTE_FN_T = Callable[[DF_T], DF_T]
COMPUTE_FN_UNBOUND_T = Callable[..., DF_T]
COMPUTE_FNTR_T = Callable[..., COMPUTE_FN_T]
ANY_COMPUTE_FN_T = COMPUTE_FN_T | COMPUTE_FN_UNBOUND_T | COMPUTE_FNTR_T


class ComputeFnArgs(TypedDict, total=False):
    df: NotRequired[DF_T]
    cfg: NotRequired[DictConfig]
    stage_cfg: NotRequired[DictConfig]
    code_modifiers: NotRequired[list[str]]
    code_metadata: NotRequired[pl.DataFrame]


SHARD_GEN_T = Generator[tuple[Path, Path], None, None]
SHARD_ITR_FNTR_T = Callable[[DictConfig], SHARD_GEN_T]


class ComputeFnType(Enum):
    """Stores the three types of compute functions that can be used in this utility.

    Attributes:
        DIRECT: A direct compute function that maps a DataFrame to a DataFrame.
        UNBOUND: An unbound compute function that needs to be called with additional parameters in addition to
            the input DataFrame, but still returns a DataFrame.
        FUNCTOR: A functor that needs to be called with non "df" args to return a direct compute function.
    """

    DIRECT = auto()
    UNBOUND = auto()
    FUNCTOR = auto()


def compute_fn_type(compute_fn: ANY_COMPUTE_FN_T) -> ComputeFnType | None:
    """Determine the type of a compute function, or return None if the type matches no compute function type.

    Returns the type of the compute function:
      - ComputeFnType.DIRECT if the compute function is a direct compute function -- i.e., if it takes only a
        single parameter named ``"df"`` and if the return annotation (if one exists) is not a
        `collections.abc.Callable` type annotation.
      - ComputeFnType.UNBOUND if the compute function is an unbound compute function -- i.e., if it takes a
        ``"df"`` parameter and at least one other parameter among the set of allowed additional or functor
        parameters, and if the return annotation (if one exists) is not a `collections.abc.Callable` type
        annotation.
      - ComputeFnType.FUNCTOR if the compute function is a functor that returns a direct compute function when
        called with the appropriate parameters -- i.e., if it takes no ``"df"`` parameter, if all parameters
        are among the allowed functor parameters, and if the return annotation (if one exists) is a
        `collections.abc.Callable` type annotation.

    Allowed functor parameters are:
      - "cfg" for the DictConfig configuration object.
      - "stage_cfg" for the DictConfig stage configuration object.
      - "code_modifiers" for a list of code modifier columns.
      - "code_metadata" for a Polars DataFrame of code metadata

    Args:
        compute_fn: The compute function to check.

    Examples:
        >>> def compute_fn(df: pl.DataFrame) -> pl.DataFrame: return df
        >>> compute_fn_type(compute_fn)
        <ComputeFnType.DIRECT: 1>
        >>> def compute_fn(df: pl.DataFrame) -> dict[Any, list[Any]]: return df.to_dict(as_series=False)
        >>> compute_fn_type(compute_fn)
        <ComputeFnType.DIRECT: 1>
        >>> def compute_fn(df: pl.DataFrame): return None
        >>> compute_fn_type(compute_fn)
        <ComputeFnType.DIRECT: 1>
        >>> def compute_fn(foo: pl.DataFrame): return None
        >>> compute_fn_type(compute_fn) is None
        True
        >>> def compute_fn(df: pl.DataFrame, cfg: DictConfig) -> pl.DataFrame: return df
        >>> compute_fn_type(compute_fn)
        <ComputeFnType.UNBOUND: 2>
        >>> def compute_fn(df: pl.DataFrame, foo: DictConfig) -> pl.DataFrame: return df
        >>> compute_fn_type(compute_fn) is None
        True
        >>> def compute_fn(df: pl.LazyFrame, cfg: DictConfig): return df
        >>> compute_fn_type(compute_fn)
        <ComputeFnType.UNBOUND: 2>
        >>> def compute_fn(cfg: DictConfig) -> Callable[[pl.DataFrame], pl.DataFrame]: return lambda df: df
        >>> compute_fn_type(compute_fn)
        <ComputeFnType.FUNCTOR: 3>
        >>> def compute_fn(cfg: DictConfig): return lambda df: df
        >>> compute_fn_type(compute_fn)
        <ComputeFnType.FUNCTOR: 3>
        >>> def compute_fn(df: pl.DataFrame, cfg: DictConfig) -> Callable[[pl.DataFrame], pl.DataFrame]:
        ...     return lambda df: df
        >>> compute_fn_type(compute_fn) is None
        True
    """
    sig = inspect.signature(compute_fn)

    allowed_params = {*ComputeFnArgs.__required_keys__, *ComputeFnArgs.__optional_keys__}
    all_params_allowed = all(param in allowed_params for param in sig.parameters.keys())
    if not all_params_allowed:
        return None

    has_df_param = "df" in sig.parameters
    has_only_df_param = has_df_param and (len(sig.parameters) == 1)
    has_return_annotation = sig.return_annotation.__name__ != "_empty"
    has_callable_return = sig.return_annotation.__name__ == "Callable"

    if has_only_df_param:
        if has_return_annotation:
            return ComputeFnType.DIRECT if not has_callable_return else None
        return ComputeFnType.DIRECT
    elif has_df_param:
        if has_return_annotation:
            return ComputeFnType.UNBOUND if not has_callable_return else None
        return ComputeFnType.UNBOUND
    else:
        if has_return_annotation:
            return ComputeFnType.FUNCTOR if has_callable_return else None
        return ComputeFnType.FUNCTOR


def identity_fn(df: Any) -> Any:
    """A "null" compute function that returns the input DataFrame as is.

    Args:
        df: The input DataFrame.

    Returns:
        The input DataFrame.

    Examples:
        >>> df = pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> (identity_fn(df) == df).select(pl.all_horizontal(pl.all().all())).item()
        True
    """

    return df


def read_and_filter_fntr(filter_expr: pl.Expr, read_fn: Callable[[Path], DF_T]) -> Callable[[Path], DF_T]:
    """Create a function that reads a DataFrame from a file and filters it based on a given expression.

    This is specified as a functor in this way to allow it to modify arbitrary other read functions for use in
    different mapreduce pipelines.

    Args:
        filter_expr: The filter expression to apply to the DataFrame.
        read_fn: The read function to use to read the DataFrame.

    Returns:
        A function that reads a DataFrame from a file and filters it based on the given expression.

    Examples:
        >>> dfs = {
        ...     "df1": pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]}),
        ...     "df2": pl.DataFrame({"a": [4, 5, 6], "b": [7, 8, 9]})
        ... }
        >>> read_fn = lambda key: dfs[key]
        >>> fn = read_and_filter_fntr((pl.col("a") % 2) == 0, read_fn)
        >>> fn("df1")
        shape: (1, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 2   ┆ 5   │
        └─────┴─────┘
        >>> fn("df2")
        shape: (2, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 4   ┆ 7   │
        │ 6   ┆ 9   │
        └─────┴─────┘
        >>> fn = read_and_filter_fntr((pl.col("b") % 2) == 0, read_fn)
        >>> fn("df1")
        shape: (2, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 1   ┆ 4   │
        │ 3   ┆ 6   │
        └─────┴─────┘
        >>> fn("df2")
        shape: (1, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 5   ┆ 8   │
        └─────┴─────┘
    """

    def read_and_filter(in_fp: Path) -> DF_T:
        return read_fn(in_fp).filter(filter_expr)

    return read_and_filter


MATCH_REVISE_KEY = "_match_revise"
MATCHER_KEY = "_matcher"
MATCH_REVISE_MODE_KEY = "_match_revise_mode"


class MatchReviseMode(StrEnum):
    """The different modes for match and revise operations.

    Future modes to be considered, match and add, multi-match and add, filter and revise, multi-filter and
    revise.

    Attributes:
        MATCH_AND_REVISE: The match and revise mode, which iterates through the list of matcher/function pairs
            and filters the input DataFrame for rows that match the matcher and applies a local compute
            function to the filtered DataFrame. The DataFrame to be matched in future iterations is restricted
            to only those rows that have not yet been matched. The unmatched dataframe at the end of the
            operation is concatenated with the outputs of all intermediate dataframes.
        MULTI_MATCH_AND_REVISE: The match and revise mode, which iterates through the list of matcher/function
            pairs and filters the input DataFrame for rows that match the matcher and applies a local compute
            function to the filtered DataFrame. The DataFrame to be matched in future iterations is the entire
            raw dataframe, including rows that have already been matched. The portion of the dataframe that
            didn't match anything on input is concatenated with the outputs of all intermediate dataframes.
    """

    MATCH_AND_REVISE = auto()
    MULTI_MATCH_AND_REVISE = auto()


def is_match_revise(stage_cfg: DictConfig) -> bool:
    """Check if the stage configuration is in a match and revise format.

    Examples:
        >>> is_match_revise(DictConfig({"_match_revise": []}))
        False
        >>> is_match_revise(DictConfig({"_match_revise": [{"_matcher": {"code": "CODE//TEMP"}}]}))
        True
        >>> is_match_revise(DictConfig({"foo": "bar"}))
        False
    """
    return bool(stage_cfg.get(MATCH_REVISE_KEY, False))


def validate_match_revise(stage_cfg: DictConfig):
    """Validate that the stage configuration is in a match and revise format.

    Examples:
        >>> validate_match_revise(DictConfig({"foo": []}))
        Traceback (most recent call last):
            ...
        ValueError: Stage configuration must contain a _match_revise key
        >>> validate_match_revise(DictConfig({"_match_revise": "foo"}))
        Traceback (most recent call last):
            ...
        ValueError: Match revise options must be a list, got <class 'str'>
        >>> validate_match_revise(DictConfig({"_match_revise": [1]}))
        Traceback (most recent call last):
            ...
        ValueError: Match revise config 0 must be a dict, got <class 'int'>
        >>> validate_match_revise(DictConfig({"_match_revise": [{"_matcher": {"foo": "bar"}}, 1]}))
        Traceback (most recent call last):
            ...
        ValueError: Match revise config 1 must be a dict, got <class 'int'>
        >>> validate_match_revise(DictConfig({"_match_revise": [{"foo": "bar"}]}))
        Traceback (most recent call last):
            ...
        ValueError: Match revise config 0 must contain a _matcher key
        >>> validate_match_revise(DictConfig({"_match_revise": [{"_matcher": {32: "bar"}}]}))
        Traceback (most recent call last):
            ...
        ValueError: Match revise config 0 must contain a valid matcher in _matcher: ...
        >>> validate_match_revise(DictConfig({"_match_revise": [{"_matcher": {"code": "CODE//TEMP"}}]}))
    """

    if MATCH_REVISE_KEY not in stage_cfg:
        raise ValueError(f"Stage configuration must contain a {MATCH_REVISE_KEY} key")

    match_revise_options = stage_cfg[MATCH_REVISE_KEY]
    if not isinstance(match_revise_options, (list, ListConfig)):
        raise ValueError(f"Match revise options must be a list, got {type(match_revise_options)}")

    for i, match_revise_cfg in enumerate(match_revise_options):
        if not isinstance(match_revise_cfg, (dict, DictConfig)):
            raise ValueError(f"Match revise config {i} must be a dict, got {type(match_revise_cfg)}")

        if MATCHER_KEY not in match_revise_cfg:
            raise ValueError(f"Match revise config {i} must contain a {MATCHER_KEY} key")

        matcher_valid, matcher_errs = is_matcher(match_revise_cfg[MATCHER_KEY])
        if not matcher_valid:
            raise ValueError(
                f"Match revise config {i} must contain a valid matcher in {MATCHER_KEY}: {matcher_errs}"
            )


def match_revise_fntr(cfg: DictConfig, stage_cfg: DictConfig, compute_fn: ANY_COMPUTE_FN_T) -> COMPUTE_FN_T:
    """A functor that creates a match & revise compute function based on the given configuration.

    Stage configurations for match & revise must be in a match and revise format. Consider the below example,
    showing the ``stage_cfg`` object in ``yaml`` format:

        ```yaml
        global_arg_1: "foo"
        _match_revise:
          - _matcher: {code: "CODE//BAR"}
            local_arg_1: "bar"
          - _matcher: {code: "CODE//BAZ"}
            local_arg_1: "baz"
        ```

    This configuration will create a match & revise compute function that will filter the input DataFrame for
    rows that match the ``CODE//BAR`` code and apply the compute function with the ``local_arg_1=bar``
    parameter, and then filter the input DataFrame for rows that match the ``CODE//BAZ`` code and apply the
    compute function with the ``local_arg_1=baz`` parameter. Both of these local compute functions will be
    applied to the input DataFrame in sequence, and the resulting DataFrames will be concatenated alongside
    any of the dataframe that matches no matcher (which will be left unmodified) and merged in a sorted way
    that respects the ``subject_id``, ``time`` ordering first, then the order of the match & revise blocks
    themselves, then the order of the rows in each match & revise block output. Each local compute function
    will also use the ``global_arg_1=foo`` parameter.

    Args:
        cfg: The DictConfig configuration object.
        stage_cfg: The DictConfig stage configuration object. This stage configuration must be in a match and
            revise format, meaning it must have a key ``"_match_revise"`` that contains a list of local match
            & revise configurations. Each local match & revise configuration must contain a key ``"_matcher"``
            which links to the matcher configuration to use to filter the input DataFrame for the local
            compute execution, and all other keys are local configuration parameters to be used in the local
            compute execution.
        compute_fn: The compute function to bind to the match & revise configuration local arguments.

    Returns:
        A function that applies the match & revise compute function to the input DataFrame.

    Raises:
        ValueError: If the stage configuration is not in a match and revise format.

    Examples:
        >>> df = pl.DataFrame({
        ...     "subject_id": [1, 1, 1, 2, 2, 2],
        ...     "time": [1, 2, 2, 1, 1, 2],
        ...     "initial_idx": [0, 1, 2, 3, 4, 5],
        ...     "code": ["FINAL", "CODE//TEMP_2", "CODE//TEMP_1", "FINAL", "CODE//TEMP_2", "CODE//TEMP_1"]
        ... })
        >>> def compute_fn(df: pl.DataFrame, stage_cfg: DictConfig) -> pl.DataFrame:
        ...     return df.with_columns(
        ...         pl.col("code").str.slice(0, len("CODE//")) +
        ...         stage_cfg.local_code_mid + "//" + stage_cfg.global_code_end
        ...     )
        >>> stage_cfg = DictConfig({
        ...     "global_code_end": "foo",
        ...     "_match_revise": [
        ...         {"_matcher": {"code": "CODE//TEMP_1"}, "local_code_mid": "bar"},
        ...         {"_matcher": {"code": "CODE//TEMP_2"}, "local_code_mid": "baz"}
        ...     ]
        ... })
        >>> cfg = DictConfig({"stage_cfg": stage_cfg})
        >>> match_revise_fn = match_revise_fntr(cfg, stage_cfg, compute_fn)
        >>> match_revise_fn(df.lazy()).collect()
        shape: (6, 4)
        ┌────────────┬──────┬─────────────┬────────────────┐
        │ subject_id ┆ time ┆ initial_idx ┆ code           │
        │ ---        ┆ ---  ┆ ---         ┆ ---            │
        │ i64        ┆ i64  ┆ i64         ┆ str            │
        ╞════════════╪══════╪═════════════╪════════════════╡
        │ 1          ┆ 1    ┆ 0           ┆ FINAL          │
        │ 1          ┆ 2    ┆ 2           ┆ CODE//bar//foo │
        │ 1          ┆ 2    ┆ 1           ┆ CODE//baz//foo │
        │ 2          ┆ 1    ┆ 4           ┆ CODE//baz//foo │
        │ 2          ┆ 1    ┆ 3           ┆ FINAL          │
        │ 2          ┆ 2    ┆ 5           ┆ CODE//bar//foo │
        └────────────┴──────┴─────────────┴────────────────┘
        >>> stage_cfg = DictConfig({
        ...     "global_code_end": "foo",
        ...     "_match_revise": [
        ...         {"_matcher": {"code": "CODE//TEMP_2"}, "local_code_mid": "bizz"},
        ...         {"_matcher": {"code": "CODE//TEMP_1"}, "local_code_mid": "foo", "global_code_end": "bar"},
        ...     ]
        ... })
        >>> cfg = DictConfig({"stage_cfg": stage_cfg})
        >>> match_revise_fn = match_revise_fntr(cfg, stage_cfg, compute_fn)
        >>> match_revise_fn(df.lazy()).collect()
        shape: (6, 4)
        ┌────────────┬──────┬─────────────┬─────────────────┐
        │ subject_id ┆ time ┆ initial_idx ┆ code            │
        │ ---        ┆ ---  ┆ ---         ┆ ---             │
        │ i64        ┆ i64  ┆ i64         ┆ str             │
        ╞════════════╪══════╪═════════════╪═════════════════╡
        │ 1          ┆ 1    ┆ 0           ┆ FINAL           │
        │ 1          ┆ 2    ┆ 1           ┆ CODE//bizz//foo │
        │ 1          ┆ 2    ┆ 2           ┆ CODE//foo//bar  │
        │ 2          ┆ 1    ┆ 4           ┆ CODE//bizz//foo │
        │ 2          ┆ 1    ┆ 3           ┆ FINAL           │
        │ 2          ┆ 2    ┆ 5           ┆ CODE//foo//bar  │
        └────────────┴──────┴─────────────┴─────────────────┘
        >>> stage_cfg = DictConfig({
        ...     "global_code_end": "foo", "_match_revise": [{"_matcher": {"missing": "CODE//TEMP_2"}}]
        ... })
        >>> cfg = DictConfig({"stage_cfg": stage_cfg})
        >>> match_revise_fn = match_revise_fntr(cfg, stage_cfg, compute_fn)
        >>> match_revise_fn(df.lazy()).collect() # doctest: +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
            ...
        ValueError: Missing needed columns {'missing'} for local matcher 0:
            [(col("missing")) == (String(CODE//TEMP_2))].all_horizontal()
        Columns available: 'code', 'initial_idx', 'subject_id', 'time'

        It will throw an error if the match and revise configuration is missing.
        >>> stage_cfg = DictConfig({"global_code_end": "foo"})
        >>> cfg = DictConfig({"stage_cfg": stage_cfg})
        >>> match_revise_fn = match_revise_fntr(cfg, stage_cfg, compute_fn)
        Traceback (most recent call last):
            ...
        ValueError: Invalid match and revise configuration...

        It does not accept invalid modes.
        >>> stage_cfg = DictConfig({
        ...     "global_code_end": "foo",
        ...     "_match_revise_mode": "foobar",
        ...     "_match_revise": [{"_matcher": {"code": "CODE//TEMP_2"}}]
        ... })
        >>> cfg = DictConfig({"stage_cfg": stage_cfg})
        >>> match_revise_fn = match_revise_fntr(cfg, stage_cfg, compute_fn)
        Traceback (most recent call last):
            ...
        ValueError: Invalid match and revise mode: foobar
    """
    try:
        validate_match_revise(stage_cfg)
    except ValueError as e:
        raise ValueError("Invalid match and revise configuration") from e

    stage_cfg = hydra.utils.instantiate(stage_cfg)

    match_revise_mode = stage_cfg.pop(MATCH_REVISE_MODE_KEY, "match_and_revise")
    if match_revise_mode not in {x.value for x in MatchReviseMode}:
        raise ValueError(f"Invalid match and revise mode: {match_revise_mode}")

    matchers_and_fns = []
    for match_revise_cfg in stage_cfg.pop(MATCH_REVISE_KEY):
        matcher, cols = matcher_to_expr(match_revise_cfg.pop(MATCHER_KEY))
        local_stage_cfg = DictConfig({**stage_cfg, **match_revise_cfg})
        local_compute_fn = bind_compute_fn(cfg, local_stage_cfg, compute_fn)

        matchers_and_fns.append((matcher, cols, local_compute_fn))

    @wraps(compute_fn)
    def match_revise_fn(df: DF_T) -> DF_T:
        matchable_df = df
        cols = set(df.collect_schema().names())

        revision_parts = []
        final_part_filters = []
        for i, (matcher_expr, need_cols, local_compute_fn) in enumerate(matchers_and_fns):
            if not need_cols.issubset(cols):
                cols_str = "', '".join(x for x in sorted(cols))
                raise ValueError(
                    f"Missing needed columns {need_cols - cols} for local matcher {i}: "
                    f"{matcher_expr}\nColumns available: '{cols_str}'"
                )
            matched_df = matchable_df.filter(matcher_expr)

            match match_revise_mode:
                case MatchReviseMode.MATCH_AND_REVISE:
                    matchable_df = matchable_df.filter(~matcher_expr)
                case MatchReviseMode.MULTI_MATCH_AND_REVISE:
                    final_part_filters.append(~matcher_expr)

            revision_parts.append(local_compute_fn(matched_df))

        if final_part_filters:
            revision_parts.append(matchable_df.filter(pl.all_horizontal(final_part_filters)))
        else:
            revision_parts.append(matchable_df)
        return pl.concat(revision_parts, how="vertical").sort([subject_id_field, "time"], maintain_order=True)

    return match_revise_fn


def bind_compute_fn(cfg: DictConfig, stage_cfg: DictConfig, compute_fn: ANY_COMPUTE_FN_T) -> COMPUTE_FN_T:
    """Bind the compute function to the appropriate parameters based on the type of the compute function.

    Args:
        cfg: The DictConfig configuration object.
        stage_cfg: The DictConfig stage configuration object. This is separated from the ``cfg`` argument
            because in some cases, such as under the match & revise paradigm, the stage config may be modified
            dynamically under different matcher conditions to yield different compute functions.
        compute_fn: The compute function to bind.

    Returns:
        The compute function bound to the appropriate parameters.

    Raises:
        ValueError: If the compute function is not a valid compute function.

    Examples:
        >>> compute_fn = bind_compute_fn(DictConfig({}), DictConfig({}), None)
        >>> compute_fn("foobar")
        'foobar'
        >>> def compute_fntr(df: pl.DataFrame, cfg: DictConfig) -> pl.DataFrame:
        ...     return df.with_columns(pl.lit(cfg.val).alias("val"))
        >>> compute_fn = bind_compute_fn(DictConfig({"val": "foo"}), None, compute_fntr)
        >>> compute_fn(pl.DataFrame({"a": [1, 2, 3]}))
        shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ val │
        │ --- ┆ --- │
        │ i64 ┆ str │
        ╞═════╪═════╡
        │ 1   ┆ foo │
        │ 2   ┆ foo │
        │ 3   ┆ foo │
        └─────┴─────┘
        >>> def compute_fntr(cfg: DictConfig) -> Callable[[pl.DataFrame], pl.DataFrame]:
        ...     return lambda df: df.with_columns(pl.lit(cfg.val).alias("val"))
        >>> compute_fn = bind_compute_fn(DictConfig({"val": "foo"}), None, compute_fntr)
        >>> compute_fn(pl.DataFrame({"a": [1, 2, 3]}))
        shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ val │
        │ --- ┆ --- │
        │ i64 ┆ str │
        ╞═════╪═════╡
        │ 1   ┆ foo │
        │ 2   ┆ foo │
        │ 3   ┆ foo │
        └─────┴─────┘
        >>> def compute_fntr(stage_cfg, cfg) -> Callable[[pl.DataFrame], pl.DataFrame]:
        ...     return lambda df: df.with_columns(
        ...         pl.lit(stage_cfg.val).alias("stage_val"), pl.lit(cfg.val).alias("cfg_val")
        ...     )
        >>> compute_fn = bind_compute_fn(DictConfig({"val": "quo"}), DictConfig({"val": "bar"}), compute_fntr)
        >>> compute_fn(pl.DataFrame({"a": [1, 2, 3]}))
        shape: (3, 3)
        ┌─────┬───────────┬─────────┐
        │ a   ┆ stage_val ┆ cfg_val │
        │ --- ┆ ---       ┆ ---     │
        │ i64 ┆ str       ┆ str     │
        ╞═════╪═══════════╪═════════╡
        │ 1   ┆ bar       ┆ quo     │
        │ 2   ┆ bar       ┆ quo     │
        │ 3   ┆ bar       ┆ quo     │
        └─────┴───────────┴─────────┘
        >>> def compute_fntr(df, code_metadata):
        ...     return df.join(code_metadata, on="a")
        >>> code_metadata_df = pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> from tempfile import TemporaryDirectory
        >>> with TemporaryDirectory() as tmpdir:
        ...     code_metadata_fp = Path(tmpdir) / "codes.parquet"
        ...     code_metadata_df.write_parquet(code_metadata_fp)
        ...     stage_cfg = DictConfig({"metadata_input_dir": tmpdir})
        ...     compute_fn = bind_compute_fn(DictConfig({}), stage_cfg, compute_fntr)
        ...     compute_fn(pl.DataFrame({"a": [1, 2, 3]}))
        shape: (3, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ 1   ┆ 4   │
        │ 2   ┆ 5   │
        │ 3   ┆ 6   │
        └─────┴─────┘
        >>> def compute_fntr(df: pl.DataFrame, cfg: DictConfig) -> Callable[[pl.DataFrame], pl.DataFrame]:
        ...     return lambda df: df
        >>> bind_compute_fn(DictConfig({}), DictConfig({}), compute_fntr)
        Traceback (most recent call last):
            ...
        ValueError: Invalid compute function
    """

    def fntr_params(compute_fn: ANY_COMPUTE_FN_T) -> ComputeFnArgs:
        compute_fn_params = inspect.signature(compute_fn).parameters
        kwargs = ComputeFnArgs()

        if "cfg" in compute_fn_params:
            kwargs["cfg"] = cfg
        if "stage_cfg" in compute_fn_params:
            kwargs["stage_cfg"] = stage_cfg
        if "code_modifiers" in compute_fn_params:
            code_modifiers = cfg.get("code_modifiers", None)
            kwargs["code_modifiers"] = code_modifiers
        if "code_metadata" in compute_fn_params:
            kwargs["code_metadata"] = pl.read_parquet(
                Path(stage_cfg.metadata_input_dir) / "codes.parquet", use_pyarrow=True
            )
        return kwargs

    if compute_fn is None:
        return identity_fn
    match compute_fn_type(compute_fn):
        case ComputeFnType.DIRECT:
            pass
        case ComputeFnType.UNBOUND:
            compute_fn = partial(compute_fn, **fntr_params(compute_fn))
        case ComputeFnType.FUNCTOR:
            compute_fn = compute_fn(**fntr_params(compute_fn))
        case _:
            raise ValueError("Invalid compute function")

    return compute_fn


def map_over(
    cfg: DictConfig,
    compute_fn: COMPUTE_FN_T | None = None,
    read_fn: Callable[[Path], DF_T] = partial(pl.scan_parquet, glob=False),
    write_fn: Callable[[DF_T, Path], None] = write_lazyframe,
    shard_iterator_fntr: SHARD_ITR_FNTR_T = shard_iterator,
) -> list[Path]:
    stage_init(cfg)

    start = datetime.now()

    train_only = cfg.stage_cfg.get("train_only", False)

    shards, includes_only_train = shard_iterator_fntr(cfg)

    if train_only:
        split_fp = Path(cfg.input_dir) / subject_splits_filepath
        if includes_only_train:
            logger.info(
                f"Processing train split only via shard prefix. Not filtering with {str(split_fp.resolve())}."
            )
        elif split_fp.exists():
            logger.info(f"Processing train split only by filtering read dfs via {str(split_fp.resolve())}")
            train_subjects = (
                pl.scan_parquet(split_fp)
                .filter(pl.col("split") == "train")
                .select(subject_id_field)
                .collect()[subject_id_field]
                .to_list()
            )
            read_fn = read_and_filter_fntr(pl.col("subject_id").is_in(train_subjects), read_fn)
        else:
            raise FileNotFoundError(
                f"Train split requested, but shard prefixes can't be used and "
                f"subject split file not found at {str(split_fp.resolve())}."
            )
    elif includes_only_train:  # pragma: no cover
        raise ValueError("All splits should be used, but shard iterator is returning only train splits?!?")

    if is_match_revise(cfg.stage_cfg):
        compute_fn = match_revise_fntr(cfg, cfg.stage_cfg, compute_fn)
    else:
        compute_fn = bind_compute_fn(cfg, cfg.stage_cfg, compute_fn)

    all_out_fps = []
    for in_fp, out_fp in shards:
        logger.info(f"Processing {str(in_fp.resolve())} into {str(out_fp.resolve())}")
        rwlock_wrap(
            in_fp,
            out_fp,
            read_fn,
            write_fn,
            compute_fn,
            do_overwrite=cfg.do_overwrite,
        )
        all_out_fps.append(out_fp)

    logger.info(f"Finished mapping in {datetime.now() - start}")
    return all_out_fps



#!/usr/bin/env python
"""A polars-to-polars transformation function for filtering subjects by sequence length."""
from collections.abc import Callable

import hydra
import polars as pl
from omegaconf import DictConfig

from MEDS_transforms import PREPROCESS_CONFIG_YAML
from MEDS_transforms.mapreduce.mapper import map_over


def occlude_outliers_fntr(
    stage_cfg: DictConfig, code_metadata: pl.LazyFrame, code_modifiers: list[str] | None = None
) -> Callable[[pl.LazyFrame], pl.LazyFrame]:
    """Filters subject events to only encompass those with a set of permissible codes.

    Args:
        df: The input DataFrame.
        stage_cfg: The configuration for the code filtering stage.

    Returns:
        The processed DataFrame.

    Examples:
        >>> code_metadata_df = pl.DataFrame({
        ...     "code":                 ["A",  "A",  "B",  "C"],
        ...     "modifier1":            [1,    2,    1,    2],
        ...     "values/n_occurrences": [3,    1,    3,    2],
        ...     "values/sum":           [0.0,  4.0,  12.0, 2.0],
        ...     "values/sum_sqd":       [27.0, 16.0, 75.0, 4.0],
        ... # for clarity: ----- mean = [0.0,  4.0,  4.0,  1.0]
        ... # for clarity: --- stddev = [3.0,  0.0,  3.0,  1.0]
        ... })
        >>> data = pl.DataFrame({
        ...     "subject_id":      [1,   1,   2,   2],
        ...     "code":            ["A", "B", "A", "C"],
        ...     "modifier1":       [1,   1,   2,   2],
        ... # for clarity: mean    [0.0, 4.0, 4.0, 1.0]
        ... # for clarity: stddev  [3.0, 3.0, 0.0, 1.0]
        ...     "numeric_value": [15., 16., 3.9, 1.0],
        ... }).lazy()
        >>> stage_cfg = DictConfig({"stddev_cutoff": 4.5})
        >>> fn = occlude_outliers_fntr(stage_cfg, code_metadata_df, ["modifier1"])
        >>> fn(data).collect()
        shape: (4, 5)
        ┌────────────┬──────┬───────────┬───────────────┬─────────────────────────┐
        │ subject_id ┆ code ┆ modifier1 ┆ numeric_value ┆ numeric_value/is_inlier │
        │ ---        ┆ ---  ┆ ---       ┆ ---           ┆ ---                     │
        │ i64        ┆ str  ┆ i64       ┆ f64           ┆ bool                    │
        ╞════════════╪══════╪═══════════╪═══════════════╪═════════════════════════╡
        │ 1          ┆ A    ┆ 1         ┆ null          ┆ false                   │
        │ 1          ┆ B    ┆ 1         ┆ 16.0          ┆ true                    │
        │ 2          ┆ A    ┆ 2         ┆ null          ┆ false                   │
        │ 2          ┆ C    ┆ 2         ┆ 1.0           ┆ true                    │
        └────────────┴──────┴───────────┴───────────────┴─────────────────────────┘

        If no standard deviation cutoff is provided, the function should return the input DataFrame unchanged:
        >>> stage_cfg = DictConfig({})
        >>> fn = occlude_outliers_fntr(stage_cfg, code_metadata_df, ["modifier1"])
        >>> fn(data).collect()
        shape: (4, 4)
        ┌────────────┬──────┬───────────┬───────────────┐
        │ subject_id ┆ code ┆ modifier1 ┆ numeric_value │
        │ ---        ┆ ---  ┆ ---       ┆ ---           │
        │ i64        ┆ str  ┆ i64       ┆ f64           │
        ╞════════════╪══════╪═══════════╪═══════════════╡
        │ 1          ┆ A    ┆ 1         ┆ 15.0          │
        │ 1          ┆ B    ┆ 1         ┆ 16.0          │
        │ 2          ┆ A    ┆ 2         ┆ 3.9           │
        │ 2          ┆ C    ┆ 2         ┆ 1.0           │
        └────────────┴──────┴───────────┴───────────────┘
    """

    stddev_cutoff = stage_cfg.get("stddev_cutoff", None)
    if stddev_cutoff is None:
        return lambda df: df

    join_cols = ["code"]
    if code_modifiers:
        join_cols.extend(code_modifiers)

    cols_to_select = ["code"]
    if code_modifiers:
        cols_to_select.extend(code_modifiers)

    mean_col = pl.col("values/sum") / pl.col("values/n_occurrences")
    stddev_col = (pl.col("values/sum_sqd") / pl.col("values/n_occurrences") - mean_col**2) ** 0.5
    if "values/mean" not in code_metadata.columns:
        cols_to_select.append(mean_col.alias("values/mean"))
    if "values/std" not in code_metadata.columns:
        cols_to_select.append(stddev_col.alias("values/std"))

    code_metadata = code_metadata.lazy().select(cols_to_select)

    def occlude_outliers_fn(df: pl.LazyFrame) -> pl.LazyFrame:
        f"""Filters out outlier numeric values from subject events.

        In particular, this function filters the DataFrame to only include numeric values that are within
        {stddev_cutoff} standard deviations of the mean for the corresponding (code, modifier) pair.
        """

        val = pl.col("numeric_value")
        mean = pl.col("values/mean")
        stddev = pl.col("values/std")
        filter_expr = (val - mean).abs() <= stddev_cutoff * stddev

        return (
            df.join(code_metadata, on=join_cols, how="left", coalesce=True)
            .with_columns(
                filter_expr.alias("numeric_value/is_inlier"),
                pl.when(filter_expr).then(pl.col("numeric_value")).alias("numeric_value"),
            )
            .drop("values/mean", "values/std")
        )

    return occlude_outliers_fn


@hydra.main(
    version_base=None, config_path=str(PREPROCESS_CONFIG_YAML.parent), config_name=PREPROCESS_CONFIG_YAML.stem
)
def main(cfg: DictConfig):
    """Occludes outliers in accordance with the aggregated code metadata.

    Note that the aggregation stage with the appropriate aggregates must be run first! See the stage configs
    for arguments.
    """

    map_over(cfg, compute_fn=occlude_outliers_fntr)


if __name__ == "__main__":  # pragma: no cover
    main()
